package obelus2.swing;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import obelus2.DNode;
import obelus2.Diagram;
import obelus2.ObelusUtilities;

// Based on the Swing Trail examples, mainly GenealogyTree
@SuppressWarnings("serial")
public class DiagramPanel extends PagePanel {
	protected static String ACTION_ADDCHILD = "addChild";
	protected static String ACTION_EDIT = "edit";

	protected Diagram diagram;
	protected DNode selectedNode;

	protected JTextField nameField; 
	protected JTree tree;
	protected InnerListener listener;
	
	protected JPopupMenu treePopup;
	protected JMenuItem addChildMenu;
	protected JMenuItem editNodeMenu;
	

	public DiagramPanel() {
		super();
		setLayout(new GridBagLayout());
		
		listener = new InnerListener();
		
		nameField = new JTextField();
		nameField.addFocusListener(listener);
		nameField.addActionListener(listener);
		
		ObelusUtilities.addGridbagLabeledProperty(this, "Name", nameField, 0);

		tree = new JTree();
		// tree.setEditable(true);
		tree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
		tree.addTreeSelectionListener(new TreeSelectionListener() { // TT-LOW Assimilate in main listener
			@Override
			public void valueChanged(TreeSelectionEvent e) {
				TreePath selectedPath = tree.getSelectionPath();
				if (selectedPath == null) { setSelectedNode(null); 
				} else { setSelectedNode((DNode)selectedPath.getLastPathComponent());
				}
			}
			
		});
		tree.setShowsRootHandles(true);
		tree.setCellRenderer(new DiagramCellRenderer());
		tree.setExpandsSelectedPaths(true);

		JScrollPane scrollPane = new JScrollPane(tree);
		add(scrollPane, new GridBagConstraints(
				0,1, // grid x y
				2,1, // span x y
				1.0,1.0, // weight x y
				GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH, // anchor, fill
				new Insets(2,2,2,2),
				0,0 // pad x y
		));
		
		tree.addMouseListener(listener);

		// Pop-up menu 
		treePopup = new JPopupMenu();

		addChildMenu = new JMenuItem("Add child");
		addChildMenu.setActionCommand(ACTION_ADDCHILD);
		addChildMenu.addActionListener(listener);
		treePopup.add(addChildMenu);
		
		editNodeMenu = new JMenuItem("Edit");
		editNodeMenu.setActionCommand(ACTION_EDIT);
		editNodeMenu.addActionListener(listener);
		treePopup.add(editNodeMenu);
		
	}
	
	/**
	 * Sets the diagram edited by the panel.
	 * Please use {@link #setEditedContent(Cloneable)} in order to support commit/cancel buttons.
	 * @param newDiagram
	 */
	public void setDiagram(Diagram newDiagram) {
		this.diagram = newDiagram;
		nameField.setText(diagram.getName());
		
		DiagramTreeModel newModel = new DiagramTreeModel(diagram.getRoot());
		tree.setModel(newModel);

	}
	
	@Override
	public void setEditedContent(Cloneable content) {
		assert content instanceof Diagram;
		super.setEditedContent(content);
		Diagram dia2 = (Diagram)getPageProperty(PROPERTY_CONTENT_EDITED);
		setDiagram(dia2);
		
	}
	
	public DNode getSelectedNode() {
		return selectedNode;
	}
	
	protected void setSelectedNode(DNode newNode) {
		if (newNode != selectedNode) {
			firePropertyChange("selectedNode", selectedNode, newNode);
			selectedNode = newNode;
		}
	}

	protected void showTreePopup (MouseEvent e) {
		// Check items that should be disabled
		if (tree.getSelectionCount() == 1) {
			addChildMenu.setEnabled(true);
			editNodeMenu.setEnabled(true);
		} else {
			addChildMenu.setEnabled(false);
			editNodeMenu.setEnabled(false);
		}
		treePopup.show(e.getComponent(), e.getX(), e.getY());
	}
	
	// Renderer - Set adequate text and tooltip
	public class DiagramCellRenderer extends DefaultTreeCellRenderer {

		@Override
		public Component getTreeCellRendererComponent(JTree tree, Object value,
				boolean sel, boolean expanded, boolean leaf, int row,
				boolean hasFocus) {

			super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
					row, hasFocus);
			if (value instanceof DNode) {
				DNode dnode = (DNode)value;
				setText(dnode.getName() + "-" + dnode.getContent());
				setToolTipText(dnode.getContent());
			}
			return this;
		}
		
	}
	

	// All the listener behavior in same class
	// Handle name editing
	// Handle right button selection and popup display
	protected class InnerListener implements MouseListener, ActionListener, FocusListener {
		//
		// Mouse Events
		@Override
		public void mouseClicked(MouseEvent e) {
			// nothing to do so far
		}
		
		@Override
		public void mousePressed(MouseEvent e) {
			if (e.getComponent() == tree) {
				if (SwingUtilities.isRightMouseButton(e)) {
					int rowSelected = tree.getRowForLocation(e.getX(), e.getY());
					tree.setSelectionRow(rowSelected);
				}
				if (e.isPopupTrigger()) {
					showTreePopup(e);
				}
			}
		}
		@Override
		public void mouseReleased(MouseEvent e) {
			if (e.getComponent() == tree) {
				if (e.isPopupTrigger()) {
					showTreePopup(e);
				}
			}
		}
		@Override
		public void mouseEntered(MouseEvent e) {
			// nothing to do
		}

		@Override
		public void mouseExited(MouseEvent e) {
			// nothing to do
			
		}
		
		
		@Override
		public void actionPerformed(ActionEvent e) {
			Object source = e.getSource();
			if (source == addChildMenu) {
				SwingMain.onDiagramPanelAddNode(DiagramPanel.this);
			} else if (source == editNodeMenu) {
				SwingMain.onDiagramPanelEditNode(DiagramPanel.this);
			} else if (source == nameField) {
				diagram.setName(nameField.getText());
				checkEdited();
			}
		}

		@Override
		public void focusGained(FocusEvent e) {
			// nothing to do?
			
		}

		@Override
		public void focusLost(FocusEvent e) {
			Object source = e.getSource();
			if (source == nameField) {
				diagram.setName(nameField.getText());
				checkEdited();
			}
			
		}

		
		
	}
}
