package obelus2.swing;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JPanel;

import obelus2.ObelusUtilities;

/**
 * Panel designed to be displayed as a page for user interaction.
 * Will contain context properties to aid the controller object (indicating to referrer page where to return, the container panel, etc).
 * the finalize() method is suggested to provide a way to release resources.
 * An observable "edited" property is available to check if the content has been edited.
 * @author Javier Aranda (javier-aranda.com)
 * CC SA BY
 */
@SuppressWarnings("serial")
public abstract class PagePanel extends JPanel {
	/** Required property: The PagePanel (or null) which spawned this page (and where to return). */
	public static final String PROPERTY_REFERRER = "referrer";
	/** Required property: The container panel (or self) containing any extra auxiliary components (e.g. buttons for closing page). */
	public static final String PROPERTY_CONTAINER = "container";
	
	/** Class of the content handled by panel (Diagram, ONode) if the panel has many. */
	public static final String PROPERTY_CONTENT_CLASS = "content-class";
	/**
	 * Property containing content (copy) under edition.
	 * @see #setEditedContent(Object)
	 */
	public static final String PROPERTY_CONTENT_EDITED = "content-edited";
	/** Property containing unmodified content instance (to tell apart if edited, and restore if needed). */
	public static final String PROPERTY_CONTENT_BACKUP = "content-backup";
	
	public static final String PROPERTY_BUTTON_EDIT_COMMIT = "button-commit";
	public static final String PROPERTY_BUTTON_EDIT_CANCEL = "button-cancel";
	public static final String PROPERTY_BUTTON_SEARCH_SELECT = "button-select";
	public static final String PROPERTY_BUTTON_SEARCH_CREATE = "button-create";
	public static final String PROPERTY_BUTTON_SEARCH_CANCEL = "button-cancel"; // Repeated. does it matter? - Different use context, no problem so far

	// TT-REDESIGN Consider future properties
	// - editingMode: editable (yes/no/custom)
	// - returnPosition: When spawning another page for nested operation, where to return (e.g. specific field affected by nested operation).

	public static final String ACTION_COMMIT = "commit";
	public static final String ACTION_CANCEL = "cancel";
	public static final String ACTION_CREATE = "create";
	public static final String ACTION_SELECT = "select";
	// TT-REDESIGN consider when editing: apply(commit without exit), revert(changes without exit).
	
	// TT-REDESIGN Consider situation with an item with subrogate elements (e.g. diagram with nodes), some modifications might be awkward to compute, some commits will be done without noticing.
	// Actually would feel much better a Done/Undo/revert approach, although that would be more complicated
	protected boolean edited = false;
	protected Map<String, Object> pageProperties = new HashMap<>();
	// Default constructor would be implicitly generated anyway
	public PagePanel() {
		
	}
	
	public Object getPageProperty(String name) {
		return pageProperties.get(name);
	}
	
	public Object setPageProperty(String name, Object newValue) {
		return pageProperties.put(name, newValue);
	}
	
	/**
	 * Takes care of setting up content edition (if page should do).<br/>
	 * The controller shall use this method, will not directly set the content-edited and content-backup properties (this concern is this method's).<br/>
	 * The content item should properly implement the  {@link Cloneable} interface.<br/>
	 * This method will set original object into content-backup property, and set a clone into content-edited (so original object might be supplied to this method as content without much concern).<br/>
	 * If an specific panel should specify an specific (bean) property for content item, it redefine this method to include assigning cloned value for editing into that property.<br/>
	 * Notice @link #getEditedContent will return the edition clone, not the original.
	 * @param content Data intended to be edited by panel. Method takes care of cloning and using copy for actual editing.
	 * @return
	 */
	public void setEditedContent(Cloneable content) {
		setPageProperty(PROPERTY_CONTENT_BACKUP, content);
		// Object#clone() defined as protected. Need reflection as a workaround
		try {
			Object copy = content.getClass().getMethod("clone").invoke(content);
			setPageProperty(PROPERTY_CONTENT_EDITED, copy);
			setEdited(false);
		} catch (IllegalAccessException | IllegalArgumentException
				| InvocationTargetException | NoSuchMethodException
				| SecurityException e) {
			ObelusUtilities.handle(e);
		}
	}
	
	/**
	 * @see #setEditedContent(Object)
	 * @return
	 */
	public Object getEditedContent() {
		return getPageProperty(PROPERTY_CONTENT_EDITED);
	}
	
	public boolean isEdited() {
		return edited;
	}
	
	protected void setEdited(boolean newEdited) {
		if (edited != newEdited){
			firePropertyChange("edited", edited, newEdited);
			edited = newEdited;
		}
	}
	
	protected void checkEdited() {
		Object original = getPageProperty(PROPERTY_CONTENT_BACKUP);
		Object edited = getPageProperty(PROPERTY_CONTENT_EDITED);
		setEdited ( ! original.equals(edited));
	}
	
	// TT-REDESIGN lambda-functions might be preferable for callback
	// TT-REDESIGN Alternatively predefined methods per action might be invoked, and use panel class to tell apart.
	public JPanel generateSearchPanelContainer(String callbackSelect, String callbackCreate, String callbackCancel) {
		SearchContainerListener listener = new SearchContainerListener(callbackSelect, callbackCreate, callbackCancel);
		ButtonsWrapperPanel wrapper = new ButtonsWrapperPanel(this);
		setPageProperty(PROPERTY_CONTAINER, wrapper);
		JButton buttonSelect = wrapper.createAndAddButton("Select", ACTION_SELECT, listener);
		buttonSelect.setEnabled(false);
		setPageProperty(PROPERTY_BUTTON_SEARCH_SELECT, buttonSelect);
		JButton buttonCreate = wrapper.createAndAddButton("Create New", ACTION_CREATE, listener);
		setPageProperty(PROPERTY_BUTTON_SEARCH_CREATE, buttonCreate);
		JButton buttonCancel = wrapper.createAndAddButton("Cancel", ACTION_CANCEL, listener);
		setPageProperty(PROPERTY_BUTTON_SEARCH_CANCEL, buttonCancel);

		return wrapper;
	}
	
	protected class SearchContainerListener implements ActionListener, PropertyChangeListener {
		protected String callbackSelect;
		protected String callbackCreate;
		protected String callbackCancel;
		public SearchContainerListener(String callbackSelect, String callbackCreate, String callbackCancel) {
			this.callbackSelect = callbackSelect;
			this.callbackCreate = callbackCreate;
			this.callbackCancel = callbackCancel;
			PagePanel.this.addPropertyChangeListener("selectedItem", this);
		}
		@Override
		public void actionPerformed(ActionEvent e) {
			String command = e.getActionCommand();
			if (ACTION_SELECT.equals(command)) { invokeCallback(callbackSelect);
			} else if (ACTION_CREATE.equals(command)) { invokeCallback(callbackCreate);
			} else {
				assert ACTION_CANCEL.equals(command);
				invokeCallback(callbackCancel);
			}
			
		}
		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			Object src = evt.getSource();
			if (src == PagePanel.this && evt.getPropertyName().equals("selectedItem")) {
				boolean enabled = evt.getNewValue() != null;
				JButton buttonSelect = (JButton)getPageProperty(PROPERTY_BUTTON_SEARCH_SELECT);
				buttonSelect.setEnabled(enabled);
			}
			
		}
	}

	public JPanel generateEditPanelContainer(String callbackCommit, String callbackCancel) {
		EditContainerListener listener = new EditContainerListener(callbackCommit, callbackCancel);
		ButtonsWrapperPanel wrapper = new ButtonsWrapperPanel(this);
		setPageProperty(PROPERTY_CONTAINER, wrapper);
		JButton buttonCommit = wrapper.createAndAddButton("Commit", ACTION_COMMIT, listener);
		buttonCommit.setEnabled(false);
		setPageProperty(PROPERTY_BUTTON_EDIT_COMMIT, buttonCommit);
		JButton buttonCancel = wrapper.createAndAddButton("Cancel", ACTION_CANCEL, listener);
		setPageProperty(PROPERTY_BUTTON_EDIT_CANCEL, buttonCancel);

		return wrapper;
	}
	
	protected class EditContainerListener implements ActionListener, PropertyChangeListener {
		protected String callbackCommit;
		protected String callbackCancel;
		public EditContainerListener(String callbackCommit, String callbackCancel) {
			this.callbackCommit = callbackCommit;
			this.callbackCancel = callbackCancel;
			PagePanel.this.addPropertyChangeListener("edited", this);
		}
		@Override
		public void actionPerformed(ActionEvent e) {
			String command = e.getActionCommand();
			if (ACTION_COMMIT.equals(command)) { invokeCallback(callbackCommit);
			} else {
				assert ACTION_CANCEL.equals(command);
				invokeCallback(callbackCancel);
			}
			
		}
		@Override
		public void propertyChange(PropertyChangeEvent evt) {
			Object src = evt.getSource();
			if (src == PagePanel.this && evt.getPropertyName().equals("edited")) {
				boolean commitEnabled = (Boolean)evt.getNewValue();
				JButton buttonCommit = (JButton)getPageProperty(PROPERTY_BUTTON_EDIT_COMMIT);
				buttonCommit.setEnabled(commitEnabled);
			}
			
		}
	}
	

	// The callbacks here are methods in SwingController.instance() with PagePanel parameter.
	protected void invokeCallback(String callbackName) {
		try {
			SwingMain.class.getMethod(callbackName, PagePanel.class).invoke(null, this);
		} catch (IllegalAccessException | IllegalArgumentException
				| InvocationTargetException | NoSuchMethodException
				| SecurityException e) {
			ObelusUtilities.handle(e);
		}
	}
	
	/**
	 * Override GC oriented finalize() method for panel finalization.
	 * Consider it shouldn't throw exception. Even if finalization was error prone, i cant think of a case where the case should do anything but just going on.
	 * Maybe some custom error logging? Just catch and use ObelusUtilities.handle method
	 */
	@Override
	public void finalize() {
		try {
			super.finalize();
		} catch (Throwable e) {
			ObelusUtilities.handle(e);
		}
	}
}
