package obelus2.swing;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultListCellRenderer;
import javax.swing.DefaultListModel;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.tree.TreePath;

import obelus2.DNode;
import obelus2.Diagram;
import obelus2.OArc;
import obelus2.ONode;
import obelus2.ObelusUtilities;

/**
 * Obelus Application corepoint for Swing GUI
 * @author Javier Aranda (javier-aranda.com)
 * CC SA BY
 */
// TT-REDESIGN More sophisticated panel handling. Application panels might have methods to be invoked when activated/deactivated/disposed-At this stage it seems it doesnt pay to do so.

public class SwingMain {
	public static final String ACTION_EXIT = "exit";
	public static final String ACTION_BROWSE_DIAGRAM = "browse-diagram";
	public static final String ACTION_BROWSE_NODE = "browse-node";
	
//	// TODO remove
//	private static SwingMain s_instance;
//	public static SwingMain instance() {
//		if (s_instance == null) s_instance = new SwingMain();
//		return s_instance;
//	}

	private static InnerListener s_listener;
	
	private static MainPanel s_mainPanel;
	
	private static JMenuBar s_menuBar;
	
	private static JMenuItem s_menuExit;
	private static JMenuItem s_menuBrowseDiagram;
	private static JMenuItem s_menuBrowseNode;
	
	
	static {
		s_mainPanel = new MainPanel();
		s_mainPanel.setPreferredSize(new Dimension(500, 500)); // TT-REDESIGN might revise size on an empirical base. 
	
		s_listener = new InnerListener();
		
		s_menuBar = new JMenuBar();
		JMenu menuBrowse = new JMenu("Browse");
		s_menuBar.add(menuBrowse);
		s_menuBrowseDiagram = new JMenuItem("Diagram");
		s_menuBrowseDiagram.setActionCommand(ACTION_BROWSE_DIAGRAM);
		s_menuBrowseDiagram.addActionListener(s_listener);
		menuBrowse.add(s_menuBrowseDiagram);
		s_menuBrowseNode = new JMenuItem("Ontology");
		s_menuBrowseNode.setActionCommand(ACTION_BROWSE_NODE);
		s_menuBrowseNode.addActionListener(s_listener);
		menuBrowse.add(s_menuBrowseNode);
		
		s_menuExit = new JMenuItem("Exit");
		s_menuExit.setActionCommand(ACTION_EXIT);
		s_menuExit.addActionListener(s_listener);
		s_menuBar.add(s_menuExit);
		
	}
	
	//
	// Search Panel
	//
	public static void openPageDiagramSearch(Object referrer, String defaultSearch) {
		SearchPanel panel = new SearchPanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, Diagram.class);
		panel.setListRenderer(new DefaultListCellRenderer()); // TT-LOW use custom renderer
		panel.setCriteria(defaultSearch);
		// TT-REDESIGN [named-callbacks]
		// Callback from buttons with named methods and fixed signature
		// Pros: Reuse same buttons panel, compact, extendible
		// Cons: Callback signature is constrained, method names might be mistyped or change.
		// Alternatives: Channel callbacks through fixed methods, inline class/lambda expression with specific code.
		panel.generateSearchPanelContainer("onSearchPanelSelect", "onSearchPanelCreate", "onSearchPanelCancel");
		
		displayPage(panel, "Search Diagram");
	}
	
	public static void openPageONodeSearch(Object referrer, String defaultSearch) {
		SearchPanel panel = new SearchPanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, ONode.class);
		panel.setListRenderer(new DefaultListCellRenderer()); // TT-LOW use custom renderer
		panel.setCriteria(defaultSearch);
		panel.generateSearchPanelContainer("onSearchPanelSelect", "onSearchPanelCreate", "onSearchPanelCancel");
		
		displayPage(panel, "Search Node");
	}
	
	/**
	 * Call from SearchPanel. A criteria is entered. Some result should be supplied.
	 * @param searchPanel
	 */
	public static void onSearchPanelCriteria(PagePanel pagePanel) {
		final SearchPanel searchPanel = (SearchPanel)pagePanel;
		searchPanel.setEnabled(false); // Disable user action while process
		final Class<?> itemClass = (Class<?>)searchPanel.getPageProperty(PagePanel.PROPERTY_CONTENT_CLASS);
		final String criteria = searchPanel.getCriteria();
		// Do the search in background, update in Event Thread when ready.
		// TT-REDESIGN Disable GUI while searching? - In this implementation is FAST (wont be much data or actual I/O).
		// TT-REDESIGN Use SwingWorker (or alternative) in an homogeneous way
		// In this implementation data access is memory bound with non-synchronized collections, so in this case SwingWorker is kept present just for the sake of example
		new SwingWorker<List<Object>,Object>() {
			@Override
			protected List<Object> doInBackground() throws Exception {
				List<Object> result = new ArrayList<>();
				
				if (Diagram.class.equals(itemClass)) {
					ObelusUtilities.getDao().searchDiagram(criteria, result);
				} else {
					assert ONode.class.equals(itemClass);
					ObelusUtilities.getDao().searchONode(criteria, result);
				}
									
				return result;
			}

			@Override
			protected void done() {
				List<Object> searchResult;
				try {
					searchResult = (List<Object>)get();
				} catch (Exception e) {
					ObelusUtilities.handle(e);
					searchResult = new ArrayList<>();
				}
				searchPanel.setSearchResult(searchResult);
				returnToPage(searchPanel);
			}

		}.execute();
		
	}
	
	// TT-REDESIGN [return-to-callbacks]
	// See [named-callbacks]
	// Handlers get over-complicated because multiple return-to cases (also hard to maintain if adding-removing cases).
	// Also the specific handling code is much separated from the "symmetric" launch point.
	// Alternatively we might follow a protocol storing the result data in the referrer page context, and invoke a "returnCallback"(or whatever name) method.
	public static void onSearchPanelSelect(PagePanel pagePanel) {
		SearchPanel searchPanel = (SearchPanel) pagePanel;
		PagePanel panelReferrer = (PagePanel)searchPanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		Object selectedItem = searchPanel.getSelectedItem();
		
		closePage(pagePanel);
		if (panelReferrer == null) {
			// Search panel opened from scratch (menu), doesn't require to update any referrer field on action
			if (selectedItem instanceof Diagram) {
				openPageDiagram(null, (Diagram)selectedItem);
			} else {
				assert selectedItem instanceof ONode;
				openPageONode(null, (ONode)selectedItem);
			}
		} else if (panelReferrer instanceof DNodePanel) {
			// We were deferred from the search linked ONode operation. Update the DNode field and return
			((DNodePanel)panelReferrer).setLinkedNode((ONode)selectedItem);
			returnToPage(panelReferrer);
		} else if (panelReferrer instanceof OArcPanel) {
			// We were looking for the target node of an arc. Set and return
			((OArcPanel)panelReferrer).setTargetNode((ONode)selectedItem);
			returnToPage(panelReferrer);
		}
		
	}

	public static void onSearchPanelCreate(PagePanel searchPanel) {
		// Search panel will be substituted by a create panel
		// The search referrer will be the create referrer
		Object panelReferrer = searchPanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		Class<?> itemClass = (Class<?>)searchPanel.getPageProperty(PagePanel.PROPERTY_CONTENT_CLASS);
		
		closePage(searchPanel);
		
		if (panelReferrer == null) {
			// Search panel opened from scratch (menu), doesn't require to update any referrer field on action
			if (itemClass.equals(Diagram.class)) {
				Diagram newDiagram = ObelusUtilities.getDao().newDiagram();
				openPageDiagram(panelReferrer, newDiagram);
			} else {
				assert itemClass.equals(ONode.class);
				ONode newONode = ObelusUtilities.getDao().newONode();
				openPageONode(null, newONode);
			}
		} else if (panelReferrer instanceof DNodePanel) {
			// We were searching the ontology node linked to a diagram node. As we create the new node before edit, we link it
			ONode newONode = ObelusUtilities.getDao().newONode();
			((DNodePanel)panelReferrer).setLinkedNode(newONode);
			openPageONode(panelReferrer, newONode);
		} else if (panelReferrer instanceof OArcPanel) {
			// We were looking for the target node of an arc.
			ONode newONode = ObelusUtilities.getDao().newONode();
			((OArcPanel)panelReferrer).setTargetNode((ONode)newONode);
			openPageONode(panelReferrer, newONode);
		}
	}

	public static void onSearchPanelCancel(PagePanel pagePanel) {
		SearchPanel searchPanel = (SearchPanel) pagePanel;
		PagePanel panelReferrer = (PagePanel)searchPanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		
		closePage(searchPanel);
		if (panelReferrer == null) {
			// Nothing to do
		} else {
			assert panelReferrer instanceof DNodePanel;
			// We were searching the ontology node linked to a diagram node. The value is not modified. Just return
			returnToPage(panelReferrer);
		}
		
	}

	
	//
	// DiagramPanel
	//
	public static void openPageDiagram(Object referrer, Diagram diagram) {
		DiagramPanel panel = new DiagramPanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
//		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, Diagram.class);
		panel.setEditedContent(diagram);
		panel.generateEditPanelContainer("onDiagramPanelCommit", "onDiagramPanelCancel");
		
		displayPage(panel, diagram.getName());
	}
	
	public static void onDiagramPanelCommit(PagePanel pagePanel) {
		DiagramPanel diaPanel = (DiagramPanel)pagePanel;
		Diagram editedDiagram = (Diagram)diaPanel.getEditedContent();
		ObelusUtilities.getDao().updateDiagram(editedDiagram);
		// diagram does not return to a referrer. just close.
		closePage(pagePanel);
	}

	public static void onDiagramPanelCancel(PagePanel pagePanel) {
		// diagram does not return to a referrer. just close.
		closePage(pagePanel);
	}
	
	public static void onDiagramPanelAddNode(PagePanel pagePanel) {
		// Adding new node in diagram tree. Create node and display for edit
		DiagramPanel diaPanel = (DiagramPanel)pagePanel;
		DNode parentNode = diaPanel.getSelectedNode();
		DNode childNode = ObelusUtilities.getDao().newDNode(parentNode);
		// Update UI, new item should be selected (TT-REDESIGN probably should be done on return, and in the DiagramPanel code)
		TreePath selectedPath = diaPanel.tree.getSelectionPath();
		((DiagramTreeModel)diaPanel.tree.getModel()).nodeStructureChanged(selectedPath);
		selectedPath = selectedPath.pathByAddingChild(childNode);
		diaPanel.tree.setSelectionPath(selectedPath);
		diaPanel.tree.scrollPathToVisible(selectedPath);
		
		openPageDNode(pagePanel, childNode);
	}
	
	public static void onDiagramPanelEditNode(PagePanel pagePanel) {
		DNode editedNode = ((DiagramPanel)pagePanel).getSelectedNode();
		openPageDNode(pagePanel, editedNode);
	}
	
	

	//
	// DNodePanel
	//
	public static void openPageDNode(Object referrer, DNode dNode) {
		DNodePanel panel = new DNodePanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
//		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, ONode.class);
		panel.setEditedContent(dNode);
		panel.generateEditPanelContainer("onDNodePanelCommit", "onDNodePanelCancel");
		
		displayPage(panel, dNode.getName());
	}

	public static void onDNodePanelCommit(PagePanel pagePanel) {
		DNodePanel nodePanel = (DNodePanel)pagePanel;
		DNode editedNode = (DNode)nodePanel.getEditedContent();
		ObelusUtilities.getDao().updateDNode(editedNode);
		
		// Referrer should be the diagram edit panel
		DiagramPanel referrer = (DiagramPanel)pagePanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		// Edited node instance is different than the one in dao. Update and notify		
		// Update UI (Note: This is shared by edit node and add child operations. annotated to be redesigned)
		TreePath selectedPath = referrer.tree.getSelectionPath();
		// selectedPath contains the old element. Must convert it and update/notify
		TreePath selectedParentPath = selectedPath.getParentPath();
		if (selectedParentPath == null) {
			selectedPath = new TreePath(editedNode);
		} else {
			selectedPath = selectedParentPath.pathByAddingChild(editedNode);
		}
		((DiagramTreeModel)referrer.tree.getModel()).valueForPathChanged(selectedPath, editedNode);
		referrer.tree.setSelectionPath(selectedPath);
		
		// Return to the referrer panel
		closePage(pagePanel);
		returnToPage(referrer);
	}

	public static void onDNodePanelCancel(PagePanel pagePanel) {
		// Similar to commit. Just return to referrer
		PagePanel referrer = (PagePanel)pagePanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		closePage(pagePanel);
		returnToPage(referrer);
	}
	
	
	public static void onDNodePanelSearchLinked(PagePanel dnodePanel) {
		// Adding new node in diagram tree. Create node and display for edit
		openPageONodeSearch(dnodePanel, "");
		
		// Will commit change at Search Panel +(select/createNew), otherwise just return to panel
	}
	
	//
	// Ontology Node
	//
	public static void openPageONode(Object referrer, ONode oNode) {
		ONodePanel panel = new ONodePanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
//		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, ONode.class);
		panel.setEditedContent(oNode);
		panel.generateEditPanelContainer("onONodePanelCommit", "onONodePanelCancel");
		
		displayPage(panel, oNode.getName());
	}
	
	public static void onONodePanelCommit(PagePanel ppanel) {
		ONodePanel nodePanel = (ONodePanel)ppanel;
		ONode editedNode = (ONode)nodePanel.getEditedContent();
		ObelusUtilities.getDao().updateONode(editedNode);
		
		// Referrer can be the dnode or the arc panel
		// Update the GUI stuff and return.
		PagePanel referrer = (PagePanel)ppanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		if (referrer instanceof DNodePanel) {
			((DNodePanel)referrer).setLinkedNode(editedNode);
		} else if (referrer instanceof OArcPanel) {
			((OArcPanel)referrer).setTargetNode(editedNode);
		}
		
		closePage(ppanel);
		if (referrer != null) {returnToPage(referrer);}
	}
	
	public static void onONodePanelCancel(PagePanel ppanel) {
		// Similar to commit. Just return to referrer
		PagePanel referrer = (PagePanel)ppanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		
		closePage(ppanel);
		if (referrer != null) {returnToPage(referrer); }
	}
	
	public static void onONodePanelAddArc(PagePanel ppanel) {
		ONodePanel nodePanel = (ONodePanel)ppanel;
		ONode editedNode = (ONode)nodePanel.getEditedContent();
		// Create and assign new Arc
		OArc newArc = ObelusUtilities.getDao().newOArc(editedNode, null, null);
		// TT-LOW Consider assigning in panel method.
		editedNode.getArcs().add(newArc.getOId()); // TT-LOW Better assign, pending redesign
		((DefaultListModel<OArc>)nodePanel.arcsList.getModel()).addElement(newArc);
		// Set the new arc (in last position) as selected one
		nodePanel.arcsList.setSelectedIndex(nodePanel.arcsList.getModel().getSize() - 1);
		
		// Go edit new Arc
		ppanel.setEnabled(false); // TT-LOW Check if other open operations disable source panel
		openPageOArc(ppanel, newArc);
		// Arc data should be updated on commit, TT-FUTURE deleted on cancel?
	}
	
	public static void onONodePanelEditArc(PagePanel ppanel) {
		OArc selectedArc = ((ONodePanel)ppanel).arcsList.getSelectedValue();
		assert selectedArc != null;
		openPageOArc(ppanel, selectedArc);
		// Arc data should be updated on commit.
	}
	
	
	//
	// Ontology Arc
	//
	public static void openPageOArc(Object referrer, OArc oArc) {
		OArcPanel panel = new OArcPanel();
		panel.setPageProperty(PagePanel.PROPERTY_REFERRER, referrer);
//		panel.setPageProperty(PagePanel.PROPERTY_CONTENT_CLASS, ONode.class);
		panel.setEditedContent(oArc);
		JPanel container = panel.generateEditPanelContainer("onOArcPanelCommit", "onOArcPanelCancel");
		
		s_mainPanel.addPanel(container, "(Arc)"); // TT-FUTURE Better description, maybe *<dstName>
	}
	
	public static void onOArcPanelCommit(PagePanel ppanel) {
		OArcPanel arcPanel = (OArcPanel)ppanel;
		OArc editedArc = (OArc)arcPanel.getEditedContent();
		ObelusUtilities.getDao().updateOArc(editedArc);
		
		// Referrer should be the ONode edit panel (from operations either "add arc" or "edit arc")
		// element to update will be the selected one.
		ONodePanel nodePanel = (ONodePanel)ppanel.getPageProperty(PagePanel.PROPERTY_REFERRER);
		((DefaultListModel<OArc>)nodePanel.arcsList.getModel()).set(
				nodePanel.arcsList.getSelectedIndex(), editedArc);
		nodePanel.checkEdited(); // TT-LOW Check if other commit/select operations update source page as they should
		
		closePage(ppanel);
		returnToPage(nodePanel);
	}
	
	public static void onOArcPanelCancel(PagePanel ppanel) {
		// As in commit callback, but don't update anything.
		closePage(ppanel);
		returnToPage((ONodePanel)ppanel.getPageProperty(PagePanel.PROPERTY_REFERRER));
	}
	
	public static void onOArcPanelSearchTarget(PagePanel ppanel) {
		// Open a Search ONode page.
		ppanel.setEnabled(false);
		openPageONodeSearch(ppanel, "");
	}
	
	
	
	//
	// End of Panel handlers
	//
	
	protected static void displayPage(PagePanel page, String title) {
		PagePanel referrer = (PagePanel)page.getPageProperty(PagePanel.PROPERTY_REFERRER);
		if (referrer != null) {
			referrer.setEnabled(false); // Avoid further interaction until we return properly.
		}
		s_mainPanel.addPanel((JPanel)page.getPageProperty(PagePanel.PROPERTY_CONTAINER), title);
	}
	
	protected static void returnToPage(PagePanel page) {
		JPanel containerPanel = (JPanel)page.getPageProperty(PagePanel.PROPERTY_CONTAINER);
		s_mainPanel.setPanelActive(containerPanel);
		page.setEnabled(true);
	}
	
	protected static void closePage(PagePanel page) {
		JPanel containerPanel = (JPanel)page.getPageProperty(PagePanel.PROPERTY_CONTAINER);
		s_mainPanel.removePanel(containerPanel);
		page.finalize();
	}
	

	
	public static class DisplayFrameRunnable implements Runnable {
		public void run() {
			//Create and set up the window.
			JFrame frame = new JFrame("Obelus02 Test");
			frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

			s_mainPanel.setOpaque(true); //content panes must be opaque
			s_mainPanel.statusBar.setText("Welcome to Obelus. Select one Action");
			frame.setContentPane(s_mainPanel);
						
			frame.setJMenuBar(s_menuBar);

			//Display the window.
			frame.setPreferredSize(new Dimension(500,500));
			frame.pack();
			frame.setVisible(true);
		}
	}
	

	protected static class InnerListener implements ActionListener {

		@Override
		public void actionPerformed(ActionEvent e) {
			String action = e.getActionCommand();
			if (action.equals(ACTION_BROWSE_DIAGRAM)) {
				openPageDiagramSearch(null, "");
			} else if (action.equals(ACTION_BROWSE_NODE)) {
				openPageONodeSearch(null, "");
			} else if (action.equals(ACTION_EXIT)) {
				System.exit(0); // TT-REDESIGN. Some civilized shutdown might be needed if real data involved.
			}
			
		}

	}

	public static void main(String[] args) {
		try {
			SwingUtilities.invokeLater(new DisplayFrameRunnable());
		} catch(Throwable e) {
			e.printStackTrace(System.err);
		}
	}


}
